"""Functions and classes to make classification easier. It adds the ability to classify and get
AUC anc accuracy in one run of the clasification. It also contains the cross validation object
StratifiedShuffleGroupSplit

Author: Arthur Dehgan"""
from itertools import permutations, product
from sklearn.base import clone, BaseEstimator
from sklearn.model_selection import LeavePGroupsOut
from sklearn.metrics import accuracy_score, roc_auc_score
import numpy as np
from numpy.random import permutation
from joblib import Parallel, delayed
from .stats import compute_pval


def _cross_val(train_index, test_index, estimator, X, y):
    """fit and predict using the given data."""
    clf = clone(estimator)
    x_train, x_test = X[train_index], X[test_index]
    y_train, y_test = y[train_index], y[test_index]
    clf.fit(x_train, y_train)
    y_pred = clf.predict(x_test)
    return y_pred, y_test


def cross_val_score(estimator, cv, X, y, groups=None, n_jobs=1):
    """Computes all crossval on the chosen estimator, cross-val and dataset.

    it can be used instead of sklearn.model_selection.cross_val_score if you want both roc_auc and
    acc in one go."""
    clf = clone(estimator)
    crossv = clone(cv, safe=False)
    results = Parallel(n_jobs=n_jobs)(
        delayed(_cross_val)(train_index, test_index, clf, X, y)
        for train_index, test_index in crossv.split(X=X, y=y, groups=groups)
    )

    AUC = not X.shape[1] > 1 and cv.n_groups > 1
    accuracy, auc_list = [], []
    for test in results:
        y_pred = test[0]
        y_test = test[1]
        acc = accuracy_score(y_test, y_pred)
        if AUC:
            auc = roc_auc_score(y_test, y_pred)
        else:
            auc = 0
        accuracy.append(acc)
        auc_list.append(auc)
    return accuracy, auc_list


def _permutations(iterable, size, limit=None):
    """Combinations Generator"""
    i = 0
    for elem in permutations(iterable, size):
        i += 1
        if limit is not None and i == limit:
            break
        yield elem


def permutation_test(estimator, cv, X, y, groups=None, n_perm=0, n_jobs=1):
    """Will do compute permutations aucs and accs."""
    acc_pscores, auc_pscores = [], []
    for _ in range(n_perm):
        perm_index = permutation(len(y))
        clf = clone(estimator)
        y_perm = y[perm_index]
        groups_perm = groups[perm_index]
        perm_acc, perm_auc = cross_val_score(clf, cv, X, y_perm, groups_perm, n_jobs)
        acc_pscores.append(np.mean(perm_acc))
        auc_pscores.append(np.mean(perm_auc))

    return acc_pscores, auc_pscores


def classification(estimator, cv, X, y, groups=None, perm=None, n_jobs=1):
    """Do a classification.

    Parameters
    ----------
    estimator : sklearn Estimator
        The estimator that will fit and be tested.
    cv : sklearn CrossValidator
        The cross-validation method that will be used to test the estimator.
    X : array
        The Data, must be of shape (n_samples x n_features).
    y : list or array
        The labels used for training and testing.
    groups : list or array, optional
        The groups for groups based cross-validations
    perm : int, optional
        The number of permutations that will be done to assert significance of the result.
        None means no permutations will be computed
    n_jobs : int, optional (default=1)
        The number of threads to use for the cross-validations. higher means faster. setting
        to -1 will use all available threads - Warning: may slow down computer. Set to -2 to
        keep a thread available for display and other tasks on the computer.

    Returns
    -------
    save : dictionnary
    The dictionnary contains all the information about the classification and the testing :
        acc_score: the mean score across all cross-validations using the
        accuracy scoring method
        auc_score: the mean score across all cross-validations using the
        roc_auc scoring method
        acc: the list of all cross-validations accuracy scores
        auc: the list of all cross-validations roc_auc scores
        if permutation is not None it also countains:
        auc_pvalue: the pvalue using roc_auc as a scoring method
        acc_pvalue: the pvalue using accuracy as a scoring method
        auc_pscores: a list of all permutation auc scores
        acc_pscores: a list of all permutation accuracy scores

    """
    y = np.asarray(y)
    X = np.asarray(X)
    if len(X) != len(y):
        raise ValueError(
            "Dimension mismatch for X and y : {}, {}".format(len(X), len(y))
        )
    if groups is not None:
        try:
            if len(y) != len(groups):
                raise ValueError("dimension mismatch for groups and y")
        except TypeError:
            print(
                "Error in classification: y or",
                "groups is not a list or similar structure",
            )
            exit()
    clf = clone(estimator)
    accuracies, aucs = cross_val_score(clf, cv, X, y, groups, n_jobs)
    acc_score = np.mean(accuracies)
    auc_score = np.mean(aucs)
    save = {
        "acc_score": [acc_score],
        "auc_score": [auc_score],
        "acc": accuracies,
        "auc": aucs,
        "n_splits": cv.get_n_splits(X, y, groups),
    }
    if perm is not None:
        acc_pscores, auc_pscores = permutation_test(clf, cv, X, y, groups, perm, n_jobs)
        acc_pvalue = compute_pval(acc_score, acc_pscores)
        auc_pvalue = compute_pval(auc_score, auc_pscores)

        save.update(
            {
                "auc_pvalue": auc_pvalue,
                "acc_pvalue": acc_pvalue,
                "auc_pscores": auc_pscores,
                "acc_pscores": acc_pscores,
            }
        )

    return save


class StratifiedShuffleGroupSplit(BaseEstimator):
    def __init__(self, n_groups, n_iter=None):
        """Pre-initialization."""
        self.n_groups = n_groups
        self.n_iter = n_iter
        self.counter = 0
        self.labels_list = []
        self.n_each = None
        self.n_labs = None
        self.labels_list = None
        self.lpgos = None
        self.indexes = None

    def _init_atributes(self, y, groups):
        """Initialization."""
        if len(y) != len(groups):
            raise Exception("Error: y and groups need to have the same length")
        if y is None:
            raise Exception("Error: y cannot be None")
        if groups is None:
            raise Exception("Error: this function requires a groups parameter")
        if self.labels_list is None:
            self.labels_list = list(set(y))
        if self.n_labs is None:
            self.n_labs = len(self.labels_list)
        assert (
            self.n_groups % self.n_labs == 0
        ), "Error: The number of groups to leave out must be a multiple of the number of classes"
        if self.n_each is None:
            self.n_each = int(self.n_groups / self.n_labs)
        if self.lpgos is None:
            lpgos, indexes = [], []
            for label in self.labels_list:
                index = np.where(y == label)[0]
                indexes.append(index)
                lpgos.append(LeavePGroupsOut(self.n_each))
            self.lpgos = lpgos
            self.indexes = np.array(indexes)

    def split(self, X, y, groups):
        """generator for splits of the data.

        Parameters
        ----------
        X : array
            The data, of shape (n_trials x n_features)
        y : list
            The labels list
        groups : list
            The groups list

        Yields
        ------
        n_splits : int
            The number of splits.
        """
        self._init_atributes(y, groups)
        y = np.asarray(y)
        groups = np.asarray(groups)
        iterators = []
        for lpgo, index in zip(self.lpgos, self.indexes):
            iterators.append(lpgo.split(index, y[index], groups[index]))
        for ite in product(*iterators):
            if self.counter == self.n_iter:
                break
            self.counter += 1
            train_idx = np.concatenate(
                [index[it[0]] for it, index in zip(ite, self.indexes)]
            )
            test_idx = np.concatenate(
                [index[it[1]] for it, index in zip(ite, self.indexes)]
            )
            yield train_idx, test_idx

    def get_n_splits(self, X, y, groups):
        """Gives the number of splits.

        Parameters
        ----------
        X : placeholder for compatibility
        y : list
            The labels list
        groups : list
            The groups list

        Returns
        -------
        n_splits : int
            The number of splits.
        """
        self._init_atributes(y, groups)
        if self.n_iter is not None:
            return self.n_iter
        groups = np.asarray(groups)
        n = 1
        for index, lpgo in zip(self.indexes, self.lpgos):
            n *= lpgo.get_n_splits(None, None, groups[index])
        return n



""" collection of feature functions to extract signal from time series

Author: Arthur Dehgan"""
import numpy as np
from scipy.signal import welch


def power_spectral_density(signal, window, overlap, fmin, fmax, fs, average=True):
    """Compute PSD using welch method

    Parameters
    ----------
    signal : array
        The signal to compute the PSD on.
    window : int
        Length of the segments.
    overlap : float
        proportion of the segment to overlap. (eg. 0.5 will make each window have a 50% overlap
        with the previous one)
    fmin : float
        The lower bound of the frequency band that will be extracted
    fmax : float
        The upper bound of the frequency band that will be extracted
    fs : float
        The sampling frequency of the signal.
    average : bool, optional (default=True)
        Will average all frequencies and will only return one average value of PSD in the
        frequency band if True. If False, the function will return the PSD for each frequency
        bin and the associates frequency values.

    Returns
    -------
    if average is True (default)
    psd : value
        The avegare PSD in the frequency band.
    if average is False:
    f, psd : list, list
        The frequencies and their corresponding amplitudes in the signal.

    """
    f, psd = welch(
        signal,
        fs=fs,
        window="hanning",
        nperseg=window,
        noverlap=int(overlap * window),
        nfft=None,
    )
    if average:
        return np.mean(psd[(f >= fmin) * (f <= fmax)])
    return f, psd



from setuptools import setup

setup(
    name="mlneurotools",
    version="0.2.4",
    description="Tools for stats and ml",
    url="http://github.com/arthurdehgan/NeuroPy-MLToolbox",
    author="Arthur Dehgan",
    author_email="arthur.dehgan@umontreal.ca",
    packages=["mlneurotools"],
    zip_safe=False,
)



"""Collection of stat functions: ttest, binom pvalues or score thresholds etc

Author: Arthur Dehgan"""
from itertools import combinations
import numpy as np
from scipy.stats import ttest_ind, ttest_rel, binom, zscore
from scipy.special import comb
from joblib import Parallel, delayed
from sys import maxsize


def compute_pval(score, perm_scores):
    """computes pvalue of an item in a distribution)"""
    n_perm = len(perm_scores)
    pvalue = (np.sum(perm_scores >= score) + 1.0) / (n_perm + 1)
    return pvalue


def is_signif(pvalue, p=0.05):
    """Tell if condition with classifier is significative.

    Returns a boolean : True if the condition is significativeat given p
    """
    answer = False
    if pvalue <= p:
        answer = True
    return answer


def rm_outliers(data, rm_outl=2):
    zs_dat = zscore(data)
    to_keep = np.where(abs(zs_dat) < rm_outl)[0]
    return data[to_keep]


def binom_pval(score, test_set, n_classes=2):
    """Computes the pvalue of your score according to the binomial law.

    Parameters
    ----------
    score : float
        The score you want to test for significance.
    test_set : array or int
        The array used for testing, or the total number of trials in the test set.
    n_classes : int, optional (default=2)
        The number of different classes in your classification problem.
    """
    if not isinstance(test_set, int):
        test_set = len(test_set)
    return binom.sf(score * test_set, test_set, 1 / n_classes)


def binomial_chance_level(test_set, p=0.05, n_classes=2):
    """Computes the chance level according to the binomial law

    Parameters
    ----------
    test_set : array or int
        The array used for testing, or the total number of trials in the test set.
    p : float, optional (default=0.05)
        The p value you want to reach.
    n_classes : int, optional (default=2)
        The number of different classes in your classification problem.

    Returns
    -------
    score : float
        The score threshold. Any score higher or equal than this score is significant with the
        pvalue that was given in input.
    """
    if not isinstance(test_set, int):
        test_set = len(test_set)
    return binom.isf(p, test_set, 1 / n_classes) / test_set


def ttest_perm(
    cond1,
    cond2,
    n_perm=0,
    correction="maxstat",
    method="indep",
    alpha=0.05,
    equal_var=False,
    two_tailed=False,
    paired=False,
    n_jobs=1,
):
    """ttest indep with permuattions and maxstat correction

    Parameters
    ----------
    cond1, cond2 : arrays
        Must be of shape (n_subject x n_eletrodes) or n_trials x n_electrodes. arrays of data
        for the independant conditions.
    n_perm : int, optionnal (default=0)
        The number of permutations to do, if n_perm = 0 then exaustive permutations will be done.
        It will take exponential time with data size.
    correction : string, optionnal (default=None)
        The choice of correction to compute pvalues, if None, no correction will be done.
        Options are 'maxstat', 'fdr', 'bonferroni', None
    method : string, optionnal (default='indep')
        Necessary only for fdr correction. Implements Benjamini/Hochberg method if 'indep' or
        Benjamini/Yekutieli if 'negcorr'.
    alpha : float, optionnal (default=0.05)
        The error rate
    equal_var : bool, optionnal (default=False)
        If the variance of the two distributions are the same. See scipy.stats.ttest_ind for more info.
    two_tailed : bool, optionnal (default=False)
        Set to True for a two-tailed ttest.
    paired : bool, optionnal (default=False)
        Set if the condition 1 and 2 are paired condition or independent.
    n_jobs : int, optionnal (default=1)
        Number of cores used to computer permutations in parallel (-1 uses all cores and will be faster)

    Returns
    -------
        tval : list
            The calculated t-statistics
        pval :
            The pvalues after permutation test and correction if selected

    """
    _check_correction(correction)

    if paired:
        tval, _ = ttest_rel(cond1, cond2)
    else:
        tval, _ = ttest_ind(cond1, cond2, equal_var=equal_var)

    perm_t = perm_test(
        cond1, cond2, n_perm, _ttest_perm, equal_var, paired=paired, n_jobs=n_jobs
    )

    pval = compute_pvalues(tval, perm_t, two_tailed, correction=correction)

    if correction in ["bonferroni", "fdr"]:
        pval = pvalues_correction(pval, correction, method)

    return tval, pval


def perm_test(cond1, cond2, n_perm, function, equal_var=False, paired=False, n_jobs=1):
    """permutation ttest.

    Parameters
    ----------
    cond1, cond2 : arrays
        Must be of shape (n_subject x n_eletrodes) or n_trials x n_electrodes. arrays of data
        for the independant conditions.
    n_perm : int
        The number of permutations to do.
    function: function
        The function to execute in parallel on the data.
    equal_var : bool, optionnal (default=False)
        If the variance of the two distributions are the same. See scipy.stats.ttest_ind for more info.
    paired : bool, optionnal (default=False)
        Set if the condition 1 and 2 are paired condition or independent.
    n_jobs : int, optionnal (default=1)
        Number of cores used to computer permutations in parallel (-1 uses all cores and will be faster)

    Returns
    -------
    perm_t :
        The list of permutation t-statistics

    """
    full_mat = np.concatenate((cond1, cond2), axis=0)
    n_samples = len(full_mat)
    perm_t = []
    n_comb = comb(n_samples, len(cond1))
    if np.isinf(n_comb):
        n_comb = maxsize
    else:
        n_comb = int(n_comb)

    if n_perm >= n_comb - 1:
        # print("All permutations will be done. n_perm={}".format(n_comb - 1))
        if n_perm == 0:
            print(
                "size of the dataset does not allow {}"
                + "permutations, instead".format(n_perm)
            )

        n_perm = n_comb
        print("All {} permutations will be done".format(n_perm))
    if n_perm > 9999:
        print("Warning: permutation number is very high : {}".format(n_perm))
        print("it might take a while to compute ttest on all permutations")

    perms_index = _combinations(range(n_samples), len(cond1), n_perm)
    perm_t = Parallel(n_jobs=n_jobs)(
        delayed(function)(full_mat, index, equal_var=equal_var, paired=paired)
        for index in perms_index
    )

    return perm_t[1:]  # the first perm is not a permutation


def compute_pvalues(tval, perm_t, two_tailed, correction):
    """computes pvalues.

    Parameters
    ----------
    tstat : list
        The computed t-statistics.
    perm_t : list
        The list of permutation t-statistics.
    two_tailed : bool
        Will activate or deactivate two-tailed ttest.
    correction : string
        The choice of correction to compute pvalues. If None, no correction will be done
        Options are 'maxstat', 'fdr', 'bonferroni', None

    Returns
    -------
    pvalues : list
        The list of pvalues after permutation test.

    """
    scaling = len(perm_t)
    perm_t = np.array(perm_t)
    pvalues = []
    tval_tocompare = tval
    if two_tailed:
        perm_t = abs(perm_t)
        tval_tocompare = abs(tval)

    if correction == "maxstat":
        perm_t = np.asarray(perm_t).max(axis=1)
        perm_t = np.array([perm_t for _ in range(len(tval))]).T

    for i, tstat in enumerate(tval_tocompare):
        p_final = 0
        compare_list = perm_t[:, i]
        for t_perm in compare_list:
            if tstat <= t_perm:
                p_final += 1 / scaling
        pvalues.append(p_final)

    pvalues = np.asarray(pvalues, dtype=np.float32)

    return pvalues


def pvalues_correction(pvalues, correction, method):
    """computes corrected pvalues from pvalues.

    Parameters
    ----------
    pvalues : list
        The list of pvalues.
    correction: string
        The choice of correction to compute pvalues. If None, no correction will be done
        Options are 'maxstat', 'fdr', 'bonferroni', None
    method : string
        Necessary only for fdr correction. Implements Benjamini/Hochberg method if 'indep' or
        Benjamini/Yekutieli if 'negcorr'.

    Returns
    -------
    pvalues : list
        The list of corrected pvalues

    """
    if correction == "bonferroni":
        pvalues *= float(np.array(pvalues).size)

    elif correction == "fdr":
        n_obs = len(pvalues)
        index_sorted_pvalues = np.argsort(pvalues)
        sorted_pvalues = pvalues[index_sorted_pvalues]
        sorted_index = index_sorted_pvalues.argsort()
        ecdf = (np.arange(n_obs) + 1) / float(n_obs)

        if method == "negcorr":
            cm = np.sum(1.0 / (np.arange(n_obs) + 1))
            ecdf /= cm
        elif method == "indep":
            pass
        else:
            raise ValueError(method, " is not a valid method option")

        raw_corrected_pvalues = sorted_pvalues / ecdf
        corrected_pvalues = np.minimum.accumulate(raw_corrected_pvalues[::-1])[::-1]
        pvalues = corrected_pvalues[sorted_index].reshape(n_obs)

    pvalues[pvalues > 1.0] = 1.0

    return pvalues


def _generate_conds(data, index):
    """

    Parameters
    ----------
    data : array
        Array of the concatenated condition data.
    index : list
        The permutation index to apply.

    Returns
    -------
    cond1, cond2 : array
        Generated array from cond1 and cond2, contains permuted values.

    """
    index = list(index)
    index_comp = list(set(range(len(data))) - set(index))
    perm_mat = np.vstack((data[index], data[index_comp]))
    cond1, cond2 = perm_mat[: len(index)], perm_mat[len(index) :]
    return cond1, cond2


def _combinations(iterable, r, limit=None):
    """combinations generator"""
    i = 0
    for e in combinations(iterable, r):
        yield e
        i += 1
        if limit is not None and i == limit:
            break


def _relative_perm(data, index, **kwargs):
    """Compute realtives changes after on a selectes permutation"""
    cond1, cond2 = _generate_conds(data, index)
    return compute_relatives(cond1, cond2, kwargs)


def _ttest_perm(data, index, equal_var, paired):
    """ttest with the permutation index"""
    cond1, cond2 = _generate_conds(data, index)
    if paired:
        return ttest_rel(cond1, cond2)[0]
    else:
        return ttest_ind(cond1, cond2, equal_var=equal_var)[0]


def _check_correction(correction):
    """Checks if correction is a correct option"""
    if correction not in ["maxstat", "bonferroni", "fdr", None]:
        raise ValueError(correction, "is not a valid correction option")


if __name__ == "__main__":
    cond1 = np.random.randn(10, 19)
    cond2 = np.random.randn(10, 19)
    # print(cond1, cond2, sep="\n")
    tval1, pval1 = ttest_perm(cond1, cond2, two_tailed=True, n_perm=1000)
    tval2, pval2 = ttest_perm(
        cond1, cond2, n_perm=1000, two_tailed=True, correction="bonferroni"
    )
    tval3, pval3 = ttest_perm(
        cond1, cond2, n_perm=1000, two_tailed=True, correction="fdr"
    )
    val, tval4, pval4 = relative_perm(cond1, cond2, n_perm=1000, two_tailed=True)
    print(pval1, pval2, pval3, pval4)
    # print(tval4, pval4, sep="\n")



"""collection of useful tools that help with setting up a pipeline

Author: Arthur Dehgan"""
import time
import functools
import numpy as np
from scipy.io import loadmat
from .stats import rm_outliers


def compute_relatives(cond1, cond2, **kwargs):
    """Computes the relative changes.

    Parameters
    ----------
    cond1, cond2 : array
        Arrays of shape (n_subject x n_eletrodes) or (n_trials x n_electrodes). The arrays of data
        for the conditions.

    Returns
    -------
    values : list
        The calculated relative changes

    """
    cond1 = np.asarray(cond1).mean(axis=0)
    cond2 = np.asarray(cond2).mean(axis=0)
    values = (cond1 - cond2) / cond2
    return values


def proper_loadmat(file_path):
    """Loads using scipy.io.loadmat, and cleans some of the metadata"""
    data = loadmat(file_path)
    clean_data = {}
    for key, value in data.items():
        if not key.startswith("__"):
            clean_data[key] = value.squeeze().tolist()
    return clean_data


def timer(func):
    """Decorator to compute time spend for the wrapped function"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        val = func(*args, **kwargs)
        time_diff = elapsed_time(start_time, time.perf_counter())
        print('"{}" executed in {}'.format(func.__name__, time_diff))
        return val

    return wrapper


def create_groups(y):
    """Generate groups from labels of shape (subject x labels)."""
    k = 0
    y = np.asarray(list(map(np.ravel, y)))
    y = np.asarray(list(map(np.asarray, y)))
    groups = []
    for sub in y:
        for _ in range(len(sub.ravel())):
            groups.append(k)
        k += 1
    groups = np.asarray(groups).ravel()
    y = np.concatenate([lab.ravel() for lab in y], axis=0).ravel()
    return y, groups


def prepare_data(
    data, labels=None, n_trials=None, rm_outl=None, random_state=0, zscore=False
):
    """prepares the data to be used in your ml pipeline.

    The function can return the given data after removing outliers, randomly selecting trials
    to balance subjects, and zscoring. It can also generate a labels and groups list.

    Parameters
    ----------
    data : list of arrays
        The data to prepare. Each entry of the list must be an array
        of shape (n_trials, n_elecs, n_samples)

    labels : list, optional
        The labels of the groups, the list must be of the same length as the data list, and
        indicate the label of each array. You need to set labels if n_trials is not set.

    n_trials : int, optional
        The number of trials to pick at random for each array in the data list. You need to set
        n_trials if labels is not set.

    rm_outl : int, optional
        The number of standard deviation away from the mean you want to keep. For example if
        rm_outl=3, then all the subjects that have a mean that is strictly superior or inferior to
        3 times the std + the mean of all subjectswill be deleted. If rm_outl is None, no outlier
        removal will be done.

    random_state : int, optional
        The random_state for the random selection of trials. Not used if n_trials is None. You
        need to change random_state if you want to bootstrap and repeat the random selection
        multiple times or it will select the same subsets of trials.

    zscore : bool, optional, default=False
        Will zscore the data for each group if set to True.

    """
    final_data = None
    if rm_outl is not None:
        data = np.asarray([rm_outliers(sub, rm_outl) for sub in data])

    sizes = [len(sub) for sub in data]
    if n_trials is not None:
        n_sub_min = min(sizes)
        if n_trials > n_sub_min:
            print(
                "can't take {} trials, taking the minimum amout {} instead".format(
                    n_trials, n_sub_min
                )
            )
            n_trials = n_sub_min

        labels = np.asarray([[lab] * n_trials for lab in labels])
    elif labels is not None:
        labels = np.asarray([[labels[i]] * size for i, size in enumerate(sizes)])
    else:
        raise Exception(
            "Error: either specify a number of trials and the "
            + "labels will be generated or give the original labels"
        )
    labels, groups = create_groups(labels)

    for submat in data:
        if submat.shape[0] == 1:
            submat = submat.ravel()
        if n_trials is not None:
            index = np.random.RandomState(random_state).choice(
                range(len(submat)), n_trials, replace=False
            )
            prep_submat = submat[index]
        else:
            prep_submat = submat

        if zscore:
            prep_submat = zscore(prep_submat)

        final_data = (
            prep_submat
            if final_data is None
            else np.concatenate((prep_submat, final_data))
        )
    return np.asarray(final_data), labels, groups


def elapsed_time(t0, t1, formating=True):
    """Time lapsed between t0 and t1.

    Returns the time (from time.time()) between t0 and t1 in a
    more readable fashion.

    Parameters
    ----------
    t0: float
        time.time() initial measure of time
        (eg. at the begining of the script)
    t1: float
        time.time() time at the end of the script
        or the execution of a function.

    """
    lapsed = abs(t1 - t0)
    if formating:
        m, h, j = 60, 3600, 24 * 3600
        nbj = lapsed // j
        nbh = (lapsed - j * nbj) // h
        nbm = (lapsed - j * nbj - h * nbh) // m
        nbs = lapsed - j * nbj - h * nbh - m * nbm
        if lapsed > j:
            formated_time = "{:.0f}j, {:.0f}h:{:.0f}m:{:.0f}s".format(
                nbj, nbh, nbm, nbs
            )
        elif lapsed > h:
            formated_time = "{:.0f}h:{:.0f}m:{:.0f}s".format(nbh, nbm, nbs)
        elif lapsed > m:
            formated_time = "{:.0f}m:{:.0f}s".format(nbm, nbs)
        else:
            formated_time = "{:.4f}s".format(nbs)
        return formated_time
    return lapsed
